import sys
import os
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'
sys.path.append(os.getcwd())
import unittest
from python_ini.ini_writer import IniWriter
from python_ini.ini_file import IniFile


def make_ini(w):
    w.comment()
    w.comment('global section')
    w.key('aA0!$%()*+-.<>?@^_{|}~', [1, 2, 3], 'all key chars')
    w.key('floats', [1.0, -.3, -.3e2], 'floating point numbers')
    w.key('booleans', [True, False, None], 'boolean values')
    w.key('strings', ['abc', 'a\\bc', "a\\'"], 'some string values')
    w.comment()
    w.comment('start a new section')
    w.section('zZ9!$%()*+-.<>?@^_{|}~', 'odd section name')
    w.key('aA0!$%()*+-.<>?@^_{|}~', [1, 2, 3], 'all key chars')
    w.key('floats', [1.0, -.3, -.3e2], 'floating point numbers')
    w.key('booleans', [True, False, None], 'boolean values')
    w.key('strings', ['abc', 'a\\\\b', "a\\'"], 'some string values')


class TestConfig(unittest.TestCase):
    """Test all configurable options."""

    def test_delimiters_1(self):
        '''Default delimiters.'''
        w = IniWriter()
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('; global section', file)
        self.assertIn('floats = ', file)
        self.assertIn('true, false, none', file)
        self.assertIn('        ; all key chars', file)
        s = ini.get_values('strings')
        # print(s)
        self.assertEqual(s[1], 'a c')
        self.assertEqual(s[2], "a'")

    def test_delimiters_2(self):
        '''#, : and space delimiters.'''
        w = IniWriter()
        w.delimiters('#', ':', ' ')
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('# global section', file)
        self.assertIn('floats: ', file)
        self.assertIn('true false none', file)
        self.assertIn('        # all key chars', file)

    def test_delimiters_3(self):
        '''#, space and space delimiters.'''
        w = IniWriter()
        w.delimiters('#', ' ', ' ')
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('# global section', file)
        self.assertIn('floats 1', file)
        self.assertIn('true false none', file)
        self.assertIn('        # all key chars', file)

    def test_booleans_1(self):
        '''Cap values for true, false, and none.'''
        w = IniWriter()
        w.booleans('TRUE', 'OFF', 'NULL')
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('TRUE, OFF, NULL', file)

    def test_booleans_2(self):
        '''Mixed cap values for true, false, and none.'''
        w = IniWriter()
        w.booleans('YeS', 'nO', 'vOiD')
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('YeS, nO, vOiD', file)

    def test_booleans_3(self):
        '''Mixed cap values for true, false, and none.'''
        w = IniWriter()
        w.booleans('True', 'False', 'None')
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('True, False, None', file)

    def test_comment_tab_1(self):
        '''Mixed cap values for true, false, and none.'''
        w = IniWriter()
        make_ini(w)
        file = w.to_string()
        # print()
        # print(file)
        ini = IniFile('m')
        ini.parse(fstr=file)
        self.assertEqual(ini.errors, None)
        self.assertIn('        ; all key chars', file)
        w.clear()
        w.comment_tab(39)
        make_ini(w)
        file = w.to_string()
        ini.parse(fstr=file)
        # print(file)
        self.assertIn('       ; all key chars', file)
        w.clear()
        w.comment_tab(30)
        make_ini(w)
        file = w.to_string()
        ini.parse(fstr=file)
        # print(file)
        self.assertIn('                              ; all key chars', file)
        w.clear()
        w.comment_tab(0)
        make_ini(w)
        ini.parse(fstr=file)
        file = w.to_string()
        # print(file)
        self.assertIn('\n; all key chars', file)

    def test_write(self):
        '''Test writing to a file.'''
        fname = 'tests/data/temp.ini'
        w = IniWriter()
        ini = IniFile('m')
        make_ini(w)
        w.write(fname)
        # print()
        # print(file)
        ini.parse(fname)
        self.assertEqual(ini.errors, None)
        os.remove(fname)


if __name__ == '__main__':
    unittest.main()



''' @file python_ini/help.py
@brief This is the INI help function.
<pre>
options:
  -h, --help            show this help message and exit
  -v, --version         display the version and copyright information

</pre>
'''
# import os
# import sys
import argparse
# sys.path.append(os.getcwd())
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'


desc = '''
This is a Python INI file parser.
'''
ep = '''NOTES: ???
'''

how_to = '''usage:
      The parser is implemented as a Python class with member functions to parse and retrieve key values.
      The writer is implemented as a Python class with member functions to create comments,
      key/value pairs and section names and write the result to a file.
      Additional functions exist to configure the format.
      See the full documentation at https://sabnf.com/docs/python-ini/index.html for complete details.

A simple Python program illustrating the parsing basics follows.

from python_ini.ini_file import IniFile
ini = IniFile()                                 # default is single-value mode
ini.parse(INI_FILE_NAME)
if(ini.errors):
    print('INI FILE ERRORS')
    print(ini.display_errors())                 # display any errors found in the INI file
keys = ini.get_keys()                           # get a list of all keys in the global section
for key in keys:
    values = ini.get_values(key)                # get the value for each key
    print(key, end=': ')
    print(values)
sections = ini.get_sections()                   # get a list of all section names
for section in sections:
    keys = ini.get_section_keys(section)        # get a list of all keys in each section
    for key in keys:
        values = ini.get_section_values(section, key)   # get the value for each section key
        print('[' + section + ']:' + key, end=' ')
        print(values)

A simple Python program illustrating the writing basics follows.

from python_ini.ini_file import IniFile
w = IniWriter()                                         # instantiate the writer
w.delimiters('#', ':', ',')                             # set all configurable values
w.booleans('TRUE', 'OFF', 'void')
w.comment_tab(30)
w.comment()                                             # generate a blank comment line
w.comment('global keys')                                # generate a comment line
w.key('Unicode', 'a\U0010ffffb', 'max Unicode character')
w.key('flags', [True, False, None], 'all "booleans"')   # generate key/value pair
w.comment()
w.comment('first section')
w.section('__SECTION__', 'this is a section')           # generate a section line
w.key('section_key', [1, 2, 3])
w.key(
    'long-key',
    ['abc\xffdef\ue000ghi\U0010ffffjkl'],
    'hex and Unicode string characters')
w.write('output.ini)                                    # write the INI file to 'output.ini'

'''


def main():
    parser = argparse.ArgumentParser(
        prog='python-ini',
        description=desc
    )
    parser.add_argument('-v', '--version',
                        help='display the version and copyright information and exit',
                        dest='version',
                        action='store_true')
    parser.add_argument('-u', '--usage',
                        help='display INI file parser usage and exit',
                        dest='usage',
                        action='store_true')
    args = parser.parse_args()

    if(args.version):
        # handle the version request
        text = 'python-ini version 1.1.0'
        text += '\nA Python INI file parser and writer'
        text += '\nCopyright (c) 2023 Lowell D. Thomas'
        print(text)
        exit()

    if(args.usage):
        # input SABNF file name required
        print(how_to)
        exit()

    print('options [--help | --version | --usage]')



import unittest
from python_ini.ini_file import IniFile


class TestGlobalSingle(unittest.TestCase):
    """Test parsing disjoint sections and keys in multi-value mode."""

    def test_multi_1(self):
        '''True flags in multi-value mode - input file name'''
        fname = 'tests/data/true_flag_globals.ini'
        # file name and values parameters are positional
        ini = IniFile('m')
        ini.parse(fname)
        keys = ini.get_keys()
        for key in keys:
            value = ini.get_values(key)
            self.assertEqual(value, [True])

    def test_multi_2(self):
        '''True flags in multi-value mode - input file handle'''
        fname = 'tests/data/true_flag_globals.ini'
        with open(fname, 'r') as fd:
            # file name and values parameters are no longer positional
            ini = IniFile('m')
            ini.parse(fhandle=fd)
            keys = ini.get_keys()
            for key in keys:
                value = ini.get_values(key)
                self.assertEqual(value, [True])

    def test_multi_3(self):
        '''True flags in multi-value mode - input string'''
        fname = 'tests/data/true_flag_globals.ini'
        with open(fname, 'r') as fd:
            input = fd.read()
            # file name and values parameters are no longer positional
            ini = IniFile('m')
            ini.parse(fstr=input)
            keys = ini.get_keys()
            for key in keys:
                value = ini.get_values(key)
                self.assertEqual(value, [True])

    def test_multi_4(self):
        '''Global, disjoint sections and disjoint keys'''
        fname = 'tests/data/disjoint.ini'
        ini = IniFile('m')
        ini.parse(fname)
        # test the global key
        keys = ini.get_keys()
        values = ini.get_values(keys[0])
        self.assertEqual(values, [1, 2, 'three'])
        sections = ini.get_sections()
        # test the first section
        keys = ini.get_section_keys(sections[0])
        values = ini.get_section_values(sections[0], keys[0])
        self.assertEqual(values, [1, 2, None])
        values = ini.get_section_values(sections[0], keys[1])
        self.assertEqual(values, [2, 3])
        values = ini.get_section_values(sections[0], keys[2])
        self.assertEqual(values, [3, 4])
        values = ini.get_section_values(sections[0], keys[3])
        self.assertEqual(values, [4, 5])
        # test the second section
        keys = ini.get_section_keys(sections[1])
        values = ini.get_section_values(sections[1], keys[0])
        self.assertEqual(values, [10, True, 50])
        values = ini.get_section_values(sections[1], keys[1])
        self.assertEqual(values, [20, False])
        values = ini.get_section_values(sections[1], keys[2])
        self.assertEqual(values, [30, 40])
        values = ini.get_section_values(sections[1], keys[3])
        self.assertEqual(values, [40, 50])


if __name__ == '__main__':
    unittest.main()



''' @file python_ini/ini_writer.py
@brief The IniWriter class for creating and writing an INI file.

The writer is configurable to specify optional delimiters, boolean values
and tab space for inline comments.
'''


class IniWriter:

    def __init__(self):
        '''Ini file writer constructor.
        Note that configurable values are set to defaults.
            - True = 'true'
            - False = 'false'
            - None = 'none'
            - ; for comment delimiter
            - = for key/value delimiter
            - , for value list delimiter
            - column 40 for tab to inline comments
        '''

        self.__SECTION = 0
        self.__KEY = 1
        self.__COMMENT = 2
        self.__TRUE = 'true'
        self.__FALSE = 'false'
        self.__NONE = 'none'
        self.__COMMENT_DELIM_SEMI = '; '
        self.__COMMENT_DELIM_HASH = '# '
        self.__COMMENT_TAB = 40
        self.__LINE_END = '\n'
        self.__KEY_DELIM_EQUALS = ' = '
        self.__KEY_DELIM_COLON = ': '
        self.__KEY_DELIM_SPACE = ' '
        self.__VALUE_DELIM_COMMA = ', '
        self.__VALUE_DELIM_SPACE = ' '
        self.__NAME_CHARS = [33, 36, 37, 38, 40, 41, 42, 43, 45,
                             46, 60, 62, 63, 64, 94, 95, 123, 124, 125, 126]
        self.clear()

    def clear(self):
        '''Initialize or reset the writer to its constructor defaults.'''
        self.errors = []
        self.__lines = []
        self.__comment_delim = self.__COMMENT_DELIM_SEMI
        self.__key_delim = self.__KEY_DELIM_EQUALS
        self.__value_delim = self.__VALUE_DELIM_COMMA
        self.__comment_tab = self.__COMMENT_TAB
        self.__true = self.__TRUE
        self.__false = self.__FALSE
        self.__none = self.__NONE

    def __validate_name(self, name):
        if(not isinstance(name, str)):
            return False
        for s in name:
            c = ord(s)
            if(c >= 48 and c <= 57):
                continue
            if(c >= 65 and c <= 90):
                continue
            if(c >= 97 and c <= 122):
                continue
            if(c in self.__NAME_CHARS):
                continue

            # invalid character in name
            return False
        return True

    def __normalize_comment(self, comment):
        if(comment is None):
            return None
        if(not isinstance(comment, str)):
            raise Exception(
                'IniWriter: invalid comment, must be string of ASCII characters 32-126', comment)
        elif(comment == '' or comment.isspace()):
            return ''
        for s in comment:
            c = ord(s)
            if(c >= 32 and c <= 126):
                continue
            raise Exception(
                'IniWriter: invalid character in comment - char code(' + str(c) + ')')
        return comment

    def __normalize_string(self, string):
        v = ''
        for s in string:
            c = ord(s)
            if(c >= 32 and c <= 126):
                v += s
            elif(c == 9):
                v += '\\t'
            elif(c == 10):
                v += '\\n'
            elif(c == 13):
                v += '\\r'
            elif(c == 39):
                v += "\\'"
            elif(c <= 0xff):
                h = str(hex(c))
                h = h[2:]
                if(len(h) == 1):
                    h = '0' + h
                v += '\\x' + h
            elif(c < 0xd800 or (c >= 0xe000 and c <= 0xffff)):
                h = str(hex(c))
                h = h[2:]
                while(len(h) < 4):
                    h = '0' + h
                v += '\\u' + h
            elif(c >= 0xd800 and c < 0xe000):
                raise Exception(
                    'IniWriter: string has Unicode surrogate value', string)
            elif(c <= 0x10ffff):
                h = str(hex(c))
                h = h[2:]
                while(len(h) < 8):
                    h = '0' + h
                v += '\\U' + h
            else:
                # Note: will probably never get here
                raise Exception(
                    'IniWriter: string has Unicode value out of range (>0x10FFFF)', string)
        return "'" + v + "'"

    def booleans(self, true=False, false=False, none=False):
        '''Sets the values to specify for boolean (and null) values.
        @param true If specified must be one of 'true', 'yes' or 'on', all case insensitive.
        @param false If specified must be one of 'false', 'no' or 'off', all case insensitive.
        @param none If specified must be one of 'none', 'null' or 'void', all case insensitive.
        @return Raises Exception on any parameter error.
        '''
        if(true):
            l = true.lower()
            if(l == 'true' or l == 'yes' or l == 'on'):
                self.__true = true
            else:
                raise Exception(
                    'IniWriter.booleans(): true argument must be one of "true", "yes" or "on", case insensitive',
                    true)
        if(false):
            l = false.lower()
            if(l == 'false' or l == 'no' or l == 'off'):
                self.__false = false
            else:
                raise Exception(
                    'IniWriter.booleans(): false argument must be one of "false", "no" or "off", case insensitive',
                    false)
        if(none):
            l = none.lower()
            if(l == 'none' or l == 'null' or l == 'void'):
                self.__none = none
            else:
                raise Exception(
                    'IniWriter.booleans(): none argument must be one of "none", "null" or "void", case insensitive',
                    none)

    def comment_tab(self, tab=None):
        '''Sets the column value for the tab to inline comments.
        @param tab The column value to begin inline comments.
        Must be an integer tab >= 0.
        @return Raises Exception on invalid tab value.
        '''
        if(tab is None):
            self.__comment_tab = self.__COMMENT_TAB
        elif(isinstance(tab, int) and (tab >= 0)):
            self.__comment_tab = tab
        else:
            raise Exception(
                'IniWriter.comment_tab(): argument must be positive integer', tab)

    def delimiters(self, comment=False, key=False, value=False):
        '''Sets the delimeter values.
        @param comment The comment delimiter. Must be semicolon(";") or hash("#").
        @param key The key/value delimiter. Must be equals("="), colon(":") or space.
        @param value The delimiter between multiple key values. Must be comma(",") or space.
        @return Raises Exception on invalid argument values.
        '''
        if(comment):
            if(comment == ';'):
                self.__comment_delim = self.__COMMENT_DELIM_SEMI
            elif(comment == '#'):
                self.__comment_delim = self.__COMMENT_DELIM_HASH
            else:
                raise Exception(
                    'IniWriter.comment_delimiters(): comment argument must be semicolon(":") or hash("#")', comment)
        if(key):
            if(key == '='):
                self.__key_delim = self.__KEY_DELIM_EQUALS
            elif(key == ':'):
                self.__key_delim = self.__KEY_DELIM_COLON
            elif(key.isspace()):
                self.__key_delim = self.__KEY_DELIM_SPACE
            else:
                raise Exception(
                    'IniWriter.key_delimiters(): key argument must be equals(=), comma(,) or space', key)
        if(value):
            if(value == ','):
                self.__value_delim = self.__VALUE_DELIM_COMMA
            elif(value.isspace()):
                self.__value_delim = self.__VALUE_DELIM_SPACE
            else:
                raise Exception(
                    'IniWriter.value_delimiters(): value argument must be comma(,) or space', value)

    def section(self, name, comment=None):
        '''Add a section line to the INI file.
        @param name The section name. Must be one or more characters from the set
            - a-zA-Z0-9!$%()*+-.<>?@^_{|}~
        @param comment An inline comment to add to the section name line.
        @returns Raises Exception on invalid name or comment.
        '''
        self.__validate_name(name)
        self.__lines.append(
            [self.__SECTION, name, self.__normalize_comment(comment)])

    def key(self, name, varg, comment=None):
        '''Add a key/value line to the INI file.
        @param name The key name. Must be one or more characters from the set
            - a-zA-Z0-9!$%()*+-.<>?@^_{|}~
        @param varg The value for the key. May be a single value or a list of values.
        Valid values are:
            - True
            - False
            - None
            - integer, positive or negative
            - floating point number, positive or negative
            - string
        @param comment An inline comment to add to the key/value line.
        @returns Raises Exception on invalid name, value or comment.
        '''
        self.__validate_name(name)
        if(isinstance(varg, list)):
            values = varg
        else:
            values = [varg]
        vlist = []
        for value in values:
            if(value is True):
                v = self.__true
            elif(value is False):
                v = self.__false
            elif(value is None):
                v = self.__none
            elif(isinstance(value, int) or isinstance(value, float)):
                v = str(value)
            elif(isinstance(value, str)):
                v = self.__normalize_string(value)
            else:
                raise Exception('IniWriter.key(): invalid value', value)
            vlist.append(v)
        self.__lines.append(
            [self.__KEY, name, vlist, self.__normalize_comment(comment)])

    def comment(self, comment=None):
        '''Add a comment line to the INI file.
        @param comment A comment string. Valid comments are:
            - None (default) adds a blank line
            - ' ' empty or all white space string adds a blank line
            - string of printing ASCII characters only, char codes 32-126
        @returns Raises Exception on invalid ncomment.
        '''
        self.__lines.append(
            [self.__COMMENT, self.__normalize_comment(comment)])

    def to_string(self):
        '''Convert all lines added with the section(), key() and comment() functions
        to a valid, formatted INI file.
        @returns The formatted INI file as a string.
        '''

        def indent(n):
            out = ''
            while(n):
                out += ' '
                n -= 1
            return out

        def add_comment(out, comment):
            lout = len(out)
            if(not comment):
                add = self.__LINE_END
            elif(lout >= self.__comment_tab):
                add = self.__LINE_END
                add += indent(self.__comment_tab)
                add += self.__comment_delim + comment
                add += self.__LINE_END
            else:
                add = indent(self.__comment_tab - lout)
                add += self.__comment_delim + comment
                add += self.__LINE_END
            return add

        # key line      = [id, name, value, comment]
        # section line  = [id, name, comment]
        # comment line  = [id, comment]
        out = ''
        for line in self.__lines:
            if(line[0] == self.__COMMENT):
                if(line[1] == ''):
                    out += self.__LINE_END
                else:
                    if(line[1] is None):
                        out += self.__LINE_END
                    else:
                        out += self.__comment_delim + line[1] + self.__LINE_END
            elif(line[0] == self.__KEY):
                out_line = line[1] + self.__key_delim
                count = 0
                for v in line[2]:
                    if(count > 0):
                        out_line += self.__value_delim
                    out_line += v
                    count += 1
                out += out_line + add_comment(out_line, line[3])
            else:
                out_line = '[' + line[1] + ']'
                out += out_line + add_comment(out_line, line[2])
        return out

    def write(self, fname):
        '''Write the formatted INI file to a file. Calls to_string().
        @param fname The file name to write the INI file to.
        @return Raises exceptions on file open or write errors.
        '''
        with open(fname, 'w') as fd:
            fd.write(self.to_string())



''' @file python_ini/parser_callbacks.py
@brief The parser call back functions for the IniFile class.
'''
import sys
import os
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'
sys.path.append(os.getcwd())
from apg_py.lib import identifiers as id


def line_end(cb):
    if(cb['state'] == id.MATCH):
        cb['user_data']['line_no'] += 1


def bad_section_line(cb):
    if(cb['state'] == id.MATCH):
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': 'bad section definition'})


def bad_value_line(cb):
    if(cb['state'] == id.MATCH):
        if(cb['user_data']['skip_bad_key']):
            cb['user_data']['skip_bad_key'] = False
        else:
            cb['user_data']['errors'].append({'line': cb['user_data']['line_no'] - 1,
                                              'message': 'bad key/value definition'})


def bad_blank_line(cb):
    if(cb['state'] == id.MATCH):
        msg = 'invalid blank line, only white space and comments allowed'
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': msg})


def hex_digit(d):
    if(d >= 48 and d <= 57):
        return d - 48
    if(d >= 65 and d <= 72):
        return d - 55
    if(d >= 97 and d <= 102):
        return d - 87
    return None


def unicode8(cb):
    cb['state'] = id.NOMATCH
    cb['phrase_length'] = 0
    d = cb['input'][cb['phrase_index']]
    if(d != 85):
        return
    c = 0
    for i in range(1, 9):
        d = cb['input'][cb['phrase_index'] + i]
        v = hex_digit(d)
        if(v is None):
            msg = 'Unicode-32 value(\\Uhhhhhhhh) has invalid hex digit - char code(' + \
                str(d) + ')'
            cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                              'message': msg})
            cb['user_data']['skip_escaped_error'] = True
            cb['user_data']['skip_bad_key'] = True
            return
        c = 16 * c + v
    if(c >= 0xd800 and c <= 0xdfff):
        msg = 'Unicode-32 value(\\Uhhhhhhhh) - surrogates not allowed - x' + \
            hex(c)
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': msg})
        cb['user_data']['skip_escaped_error'] = True
        cb['user_data']['skip_bad_key'] = True
    elif(c > 0x10ffff):
        msg = 'Unicode-32 value(\\Uhhhhhhhh) - value out of range - ' + \
            hex(c)
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': msg})
        cb['user_data']['skip_escaped_error'] = True
        cb['user_data']['skip_bad_key'] = True
    else:
        # successful match
        cb['state'] = id.MATCH
        cb['phrase_length'] = 9


def unicode4(cb):
    cb['state'] = id.NOMATCH
    cb['phrase_length'] = 0
    d = cb['input'][cb['phrase_index']]
    if(d != 117):
        return
    c = 0
    for i in range(1, 5):
        d = cb['input'][cb['phrase_index'] + i]
        v = hex_digit(d)
        if(v is None):
            msg = 'Unicode-16 value(\\uhhhh) has invalid hex digit - char code(' + \
                str(d) + ')'
            cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                              'message': msg})
            cb['user_data']['skip_escaped_error'] = True
            cb['user_data']['skip_bad_key'] = True
            return
        # validate the value
        c = 16 * c + v
    if(c >= 0xd800 and c <= 0xdfff):
        msg = 'Unicode-16 value(\\uhhhh) - surrogates not allowed - ' + \
            hex(c)
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': msg})
        cb['user_data']['skip_bad_key'] = True
        cb['user_data']['skip_escaped_error'] = True
    else:
        # successful match
        cb['state'] = id.MATCH
        cb['phrase_length'] = 5


def hexadecimal(cb):
    cb['state'] = id.NOMATCH
    cb['phrase_length'] = 0
    d = cb['input'][cb['phrase_index']]
    if(d != 120):
        return
    for i in range(1, 3):
        d = cb['input'][cb['phrase_index'] + i]
        v = hex_digit(d)
        if(v is None):
            msg = 'hexadecimal value(\\xhh) has invalid hex digit - char code(' + \
                str(d) + ')'
            cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                              'message': msg})
            cb['user_data']['skip_bad_key'] = True
            cb['user_data']['skip_escaped_error'] = True
            return
    # successful match
    cb['state'] = id.MATCH
    cb['phrase_length'] = 3


def escaped_error(cb):
    cb['state'] = id.NOMATCH
    cb['phrase_length'] = 0
    if(cb['user_data']['skip_escaped_error']):
        cb['user_data']['skip_escaped_error'] = False
    else:
        d = cb['input'][cb['phrase_index']]
        msg = 'unrecognized escaped character(\\c) - char code c(' + \
            str(d) + ')'
        cb['user_data']['errors'].append({'line': cb['user_data']['line_no'],
                                          'message': msg})
        cb['user_data']['skip_bad_key'] = True



''' @file python_ini/ini_file.py
@brief The IniFile class for parsing an INI file into section/key/values.
@dir docs @brief Supplemental documentation files.
@dir python_ini @brief The IniFile class.
@dir examples @brief Several examples of using the IniFile class.
'''
import sys
import os
import copy
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'
sys.path.append(os.getcwd())
from apg_py.lib import utilities as utils
from apg_py.lib.parser import Parser
from apg_py.lib.trace import Trace
from apg_py.lib.ast import Ast
import python_ini.grammar as grammar
import python_ini.parser_callbacks as pcb
import python_ini.ast_callbacks as acb

# Note: The SABNF syntax for the ini file parser is in grammar.abnf.
# The grammar object, grammar.py was generated with (assuming PyPI installation of apg-py)
# python3 apg-py -i python_ini/grammar.abnf


class IniFile:

    def __init__(self, values='s', debug=False):
        '''Ini file parser constructor.
        @param values Determines whether the "getter" functions", get_values() and get_section_values(),
        return a single value or a list of one or more values.
            - 's' (default) Only a single value is returned for a given key.
                If more than one value is defined for the key, only the
                last value is returned.
            - 'm' Keys are captured as a list of one or more values.
                If the same key appears more than once within a given section,
                the value(s) are appended to the previous list.
        @param debug If True, a trace of the parse is printed to stdout.
        '''
        if(not (values == 'm' or values == 's')):
            msg = 'values must be "s" (single-valued) or "m" (multi-valued)'
            raise Exception(msg, values)
        self.errors = None
        self.__parser = Parser(grammar)
        if(debug):
            Trace(self.__parser, mode='xc')
        self.__parser.add_callbacks({'line-end': pcb.line_end})
        self.__parser.add_callbacks({'bad-section-line': pcb.bad_section_line})
        self.__parser.add_callbacks({'bad-value-line': pcb.bad_value_line})
        self.__parser.add_callbacks({'bad-blank-line': pcb.bad_blank_line})
        self.__parser.add_callbacks({'u_unicode8': pcb.unicode8})
        self.__parser.add_callbacks({'u_unicode4': pcb.unicode4})
        self.__parser.add_callbacks({'u_hexadecimal': pcb.hexadecimal})
        self.__parser.add_callbacks({'u_escaped-error': pcb.escaped_error})
        self.__ast = Ast(self.__parser)
        self.__ast.add_callback('section-name', acb.section_name)
        self.__ast.add_callback('key-name', acb.key_name)
        self.__ast.add_callback('value', acb.value)
        self.__ast.add_callback('d-quoted-value', acb.d_value)
        self.__ast.add_callback('s-quoted-value', acb.s_value)
        self.__ast.add_callback('string', acb.string_value)
        self.__ast.add_callback('true', acb.true_value)
        self.__ast.add_callback('false', acb.false_value)
        self.__ast.add_callback('null', acb.null_value)
        self.__ast.add_callback('int', acb.int_value)
        self.__ast.add_callback('float', acb.float_value)
        self.__data = {}
        self.__data['values'] = values
        self.__data['current_section'] = None
        self.__data['current_key'] = None
        self.__data['global'] = {}
        self.__data['sections'] = {}

    def parse(self, fname=None, fhandle=None, fstr=None):
        '''Parse the input INI file.
        Parses the ini file, input as a file name, file handle or string.
        At least one of fname, fhandle or fstr must be supplied.
        If more than one, the first non-None value of fname, fhandle or fstr in that
        order is accepted. If none are supplied an Exception is raised.
        @param fname The name of the ini file to parse.
        @param fhandle A handle to an open ini file.
        @param fstr The ini file as a string.
        '''
        self.errors = None
        if(fname):
            with open(fname, 'r') as fd:
                input = fd.read()
        elif(fhandle):
            input = fhandle.read()
        elif(fstr):
            input = fstr
        else:
            msg = 'no input supplied, must supply fname, fhandle or fstr'
            self.errors = msg
            raise Exception(msg)
        data = {}
        data['line_no'] = 1
        data['errors'] = []
        data['skip_escaped_error'] = False
        data['skip_bad_key'] = False
        result = self.__parser.parse(
            utils.string_to_tuple(input), user_data=data)
        if(not result.success):
            # ABNF syntax is designed so that this should never happen
            msg = 'internal error - parser failed'
            msg += '\nuse IniFile(debug=True) for a trace of the parser'
            raise Exception(msg)
        if(len(data['errors'])):
            # display errors
            msg = ''
            for error in data['errors']:
                msg += '{'
                count = 0
                for key, value in error.items():
                    if(count > 0):
                        msg += ', '
                    msg += str(key) + ': ' + str(value)
                    count += 1
                msg += '}\n'
            self.errors = msg
            # raise Exception('ini file syntax errors found')
        self.__data['current_section'] = None
        self.__data['current_key'] = None
        self.__data['global'] = {}
        self.__data['sections'] = {}
        self.__ast.translate(self.__data)

    def display_errors(self):
        '''Converts any errors found in the INI file syntax to
        a human-readable ASCII string.
        @returns Returns the errors display or None if none are found.
        '''
        display = None
        if(self.errors):
            display = copy.copy(self.errors)
        return display

    def get_keys(self):
        '''Get a list of the global key names.
        @returns Returns a list, possibly empty, of global key names.
        '''
        keys = []
        for key, value in self.__data['global'].items():
            keys.append(key)
        return keys

    def get_values(self, key, default=None):
        '''Get the list of values for a global key name.
        @param key The name of the key to get values for.
        @param default The value to return if the key name is not found.
        The default value ignores the single- or multi-valued mode
        and is returned exactly as given to the user.
        It is the user's responsibility to supply a default that
        makes sense to his application.
        @returns Returns
            - default if the key name is not found
            - True if the key name list is empty (a default "true" flag)
            - the full list of values if the values switch, "m"(multiple), is set in the constructor
            - the last name in the list if the values switch, "s"(single), is set in the constructor
        '''
        values = self.__data['global'].get(key)
        if(values is None):
            return default
        if(self.__data['values'] == 'm'):
            if(len(values) == 0):
                return [True]
            return values
        if(len(values) == 0):
            return True
        return values[len(values) - 1]

    def get_sections(self):
        '''Get a list of the section names.
        @returns Returns a list, possibly empty, of section names.
        '''
        sections = []
        for key, value in self.__data['sections'].items():
            sections.append(key)
        return sections

    def get_section_keys(self, section):
        '''Get a list of key names in the named section.
        @param section The section name to find the key names in .
        @returns Returns a list, possibly empty, of key names.
        '''
        keys = []
        if(self.__data['sections'].get(section)):
            for key, value in self.__data['sections'][section].items():
                keys.append(key)
        return keys

    def get_section_values(self, section, key, default=None):
        '''Get a list of values for the named key in the named section.
        @param section The section name to find the key in.
        @param key The key name to find the list of values for.
        @param default The value to return if the section or key name is not found.
        The default value ignores the single- or multi-valued mode
        and is returned exactly as given to the user.
        It is the user's responsibility to supply a default that
        makes sense to his application.
        @returns Returns
            - default if the section or key name is not found
            - True if the key name list is empty (a default "true" flag)
            - the full list of values if the multiple values switch, "m", is set in the constructor
            - the last name in the list if the single value switch, "s", is set in the constructor
        '''
        if(self.__data['sections'].get(section) is None):
            return default
        values = self.__data['sections'][section].get(key)
        if(values is None):
            return default
        if(self.__data['values'] == 'm'):
            if(len(values) == 0):
                return [True]
            return values
        # to get here self.__data['values'] == 's'
        if(len(values) == 0):
            return True
        return values[len(values) - 1]



import unittest
import re
from python_ini.ini_file import IniFile


class TestHex(unittest.TestCase):
    """Test hexadecimal and Unicode values."""

    def test_hex_1(self):
        '''Good and Bad single hex values.'''
        ini_str = '''good = \\x00, \\x09 \\xff
one_bad = \\x00, \\xfg \\x10
all_bad = \\xxy, \\xyz \\xfy
bad = \\xa
'''
        ini = IniFile()
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        good = ini.get_values('good')
        bad = ini.get_values('one_bad')
        found = re.search(r'line: 2.*char code\(103\)', display)
        self.assertTrue(found)
        found = re.search(r'line: 3.*char code\(120\)', display)
        self.assertTrue(found)
        found = re.search(r'line: 4.*char code\(10\)', display)
        self.assertTrue(found)
        self.assertEqual(ord(good), 0xff)
        self.assertEqual(bad, None)

    def test_hex_2(self):
        '''Good and Bad multi-value hex values.'''
        ini_str = '''good = \\x00, \\x09 \\xff
one_bad = \\x00, \\xfg \\x10
all_bad = \\xxy, \\xyz \\xfy
bad = \\xa
'''
        ini = IniFile('m')
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        good = ini.get_values('good')
        self.assertIn('line: 2', display)
        self.assertIn('char code(103)', display)
        self.assertIn('line: 3', display)
        self.assertIn('char code(120)', display)
        self.assertIn('\x00', good)
        self.assertIn('\x09', good)
        self.assertIn('\xff', good)

    def test_hex_3(self):
        '''Good and Bad Unicode-16 single values.'''
        ini_str = '''good = \\u0000, \\u0009 \\ud7ff
one_bad = \\u0000, \\u0009 \\ud7fg
surrogate = \\ud800
'''
        ini = IniFile('s')
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        good = ini.get_values('good')
        self.assertEqual(ord(good), 0xd7ff)
        bad = ini.get_values('one_bad')
        self.assertFalse(bad)
        found = re.search(
            r'line: 2.*Unicode-16 value.*invalid hex digit.*char code\(103\)',
            display)
        self.assertTrue(found)
        found = re.search(
            r'line: 3.*Unicode-16 value.*surrogates not allowed.*0xd800',
            display)
        self.assertTrue(found)

    def test_hex_4(self):
        '''Good and Bad Unicode-16 multi values.'''
        ini_str = '''good = \\u0000, \\u0009 \\ud7ff
one_bad = \\u0000, \\u0009 \\ud7fg
surrogate = \\ud800
'''
        ini = IniFile('m')
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        good = ini.get_values('good')
        self.assertIn('\u0000', good)
        self.assertIn('\u0009', good)
        self.assertIn('\ud7ff', good)
        bad = ini.get_values('one_bad')
        self.assertFalse(bad)
        found = re.search(
            r'line: 2.*Unicode-16 value.*invalid hex digit.*char code\(103\)',
            display)
        self.assertTrue(found)
        found = re.search(
            r'line: 3.*Unicode-16 value.*surrogates not allowed.*0xd800',
            display)
        self.assertTrue(found)

    def test_hex_5(self):
        '''Good and Bad Unicode-32 single values.'''
        ini_str = '''good = \\U00000000, \\U00000009 \\U0010ffff
one_bad = \\U00000000, \\U00000009 \\U0000d7fg
surrogate = \\U0000d800
range = \\U00110000
'''
        ini = IniFile('s')
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        good = ini.get_values('good')
        # print('good = ', end='')
        # print(good)
        self.assertEqual(ord(good), 0x10ffff)
        bad = ini.get_values('surrogate')
        self.assertFalse(bad)
        bad = ini.get_values('range')
        # print('bad', end=' = ')
        # print(bad)
        self.assertFalse(bad)
        found = re.search(
            r'line: 2.*Unicode-32 value.*invalid hex digit.*char code\(103\)',
            display)
        self.assertTrue(found)
        found = re.search(
            r'line: 3.*Unicode-32 value.*surrogates not allowed.*0xd800',
            display)
        self.assertTrue(found)
        found = re.search(
            r'line: 4.*Unicode-32 value.*value out of range.*0x110000',
            display)
        self.assertTrue(found)

    def test_hex_6(self):
        '''Bad escaped characters.'''
        ini_str = '''escaped = \\00000000
bad = \\U00000000, \\X00000009 \\U0000d7fg
'''
        ini = IniFile('s')
        ini.parse(fstr=ini_str)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print()
        # print(display)
        found = re.search(
            r'line: 1.*unrecognized escaped character.*char code c\(48\)',
            display)
        self.assertTrue(found)
        found = re.search(
            r'line: 2.*unrecognized escaped character.*char code c\(88\)',
            display)
        self.assertTrue(found)


if __name__ == '__main__':
    unittest.main()



import unittest
from python_ini.ini_file import IniFile


class TestGlobalSingle(unittest.TestCase):
    """Test the use of default values."""

    def test_default_1(self):
        '''Global default keys = multi.'''
        fname = 'tests/data/disjoint.ini'
        ini = IniFile('m')
        ini.parse(fname)
        value = ini.get_values('notkey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_values('notkey', [1, 2])
        self.assertEqual(value, [1, 2])

    def test_default_2(self):
        '''Global default keys - single.'''
        fname = 'tests/data/disjoint.ini'
        ini = IniFile()
        ini.parse(fname)
        value = ini.get_values('notkey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_values('notkey', [1, 2])
        self.assertEqual(value, [1, 2])
        ini = IniFile('m')
        ini.parse(fname)
        value = ini.get_values('notkey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_values('notkey', [1, 2])
        self.assertEqual(value, [1, 2])

    def test_default_3(self):
        '''No section name.'''
        fname = 'tests/data/disjoint.ini'
        ini = IniFile()
        ini.parse(fname)
        value = ini.get_section_values('nosection', 'nokey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_section_values('nosection', 'nokey', [1, 2])
        self.assertEqual(value, [1, 2])
        ini = IniFile('m')
        ini.parse(fname)
        value = ini.get_section_values('nosection', 'nokey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_section_values('nosection', 'nokey', [1, 2])
        self.assertEqual(value, [1, 2])

    def test_default_4(self):
        '''No section key name.'''
        fname = 'tests/data/disjoint.ini'
        ini = IniFile()
        ini.parse(fname)
        value = ini.get_section_values('_SECTION_', 'nokey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_section_values('_SECTION_', 'nokey', [1, 2])
        self.assertEqual(value, [1, 2])
        ini = IniFile('m')
        ini.parse(fname)
        value = ini.get_section_values('_SECTION_', 'nokey', 'default-value')
        self.assertEqual(value, 'default-value')
        value = ini.get_section_values('_SECTION_', 'nokey', [1, 2])
        self.assertEqual(value, [1, 2])


if __name__ == '__main__':
    unittest.main()



import unittest
from python_ini.ini_file import IniFile


class TestStrings(unittest.TestCase):
    """Test string values."""

    def test_string_1(self):
        '''Bad string.'''
        fname = 'tests/data/interior_quotes.ini'
        ini = IniFile()
        ini.parse(fname)
        self.assertTrue(ini.errors)
        display = ini.display_errors()
        # print(display)
        self.assertTrue('bad key/value' in display)
        self.assertTrue('bad section' in display)

    def test_string_2(self):
        '''Bad string with good key/value line.'''
        fname = 'tests/data/interior_quotes.ini'
        ini = IniFile()
        ini.parse(fname)
        self.assertTrue(ini.errors)
        value = ini.get_values('key')
        self.assertEqual(value, 'value')


if __name__ == '__main__':
    unittest.main()



''' @file python_ini/ast_callbacks.py
@brief The AST call back functions for translation of the IniFile parser result.
'''
# import sys
# import os
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'
# sys.path.append(os.getcwd())
from apg_py.lib import identifiers as id
from apg_py.lib import utilities as utils


def section_name(state, input, index, length, data):
    if(state == id.SEM_PRE):
        name = utils.tuple_to_string(input[index:index + length])
        if(name != data['current_section']):
            data['current_section'] = name
        if(not data['sections'].get(name)):
            data['sections'][name] = {}


def key_name(state, input, index, length, data):
    if(state == id.SEM_PRE):
        name = utils.tuple_to_string(input[index:index + length])
        if(data['current_section']):
            section = data['sections'][data['current_section']]
        else:
            section = data['global']
        key = section.get(name)
        if(not key):
            section[name] = []
        data['current_key'] = name


def value(state, input, index, length, data):
    if(state == id.SEM_POST):
        if(data['current_section']):
            section = data['sections'][data['current_section']]
        else:
            section = data['global']
        section[data['current_key']].append(data['value'])


def hex_digit(d):
    if(d >= 48 and d <= 57):
        return d - 48
    if(d >= 65 and d <= 72):
        return d - 55
    if(d >= 97 and d <= 102):
        return d - 87
    raise Exception('bad hex digit', d)


def string_eval(input):
    value = ''
    skip = 0
    for i in range(len(input)):
        if(skip):
            skip -= 1
        else:
            if(input[i] == 92):
                if(input[i + 1] == 120):
                    c = hex_digit(input[i + 2])
                    c = 16 * c + hex_digit(input[i + 3])
                    value += chr(c)
                    skip = 3
                elif(input[i + 1] == 117):
                    c = hex_digit(input[i + 2])
                    c = 16 * c + hex_digit(input[i + 3])
                    c = 16 * c + hex_digit(input[i + 4])
                    c = 16 * c + hex_digit(input[i + 5])
                    value += chr(c)
                    skip = 5
                elif(input[i + 1] == 85):
                    c = hex_digit(input[i + 2])
                    c = 16 * c + hex_digit(input[i + 3])
                    c = 16 * c + hex_digit(input[i + 4])
                    c = 16 * c + hex_digit(input[i + 5])
                    c = 16 * c + hex_digit(input[i + 6])
                    c = 16 * c + hex_digit(input[i + 7])
                    c = 16 * c + hex_digit(input[i + 8])
                    c = 16 * c + hex_digit(input[i + 9])
                    value += chr(c)
                    skip = 9
                else:
                    if(input[i + 1] == 116):
                        value += chr(0x09)
                    elif(input[i + 1] == 114):
                        value += chr(0x0D)
                    elif(input[i + 1] == 110):
                        value += chr(0x0A)
                    elif(input[i + 1] == 98):
                        value += chr(0x20)
                    else:
                        value += chr(input[i + 1])
                    skip = 1
            else:
                value += chr(input[i])
    return value


def d_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = string_eval(input[index:index + length])


def s_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = string_eval(input[index:index + length])


def string_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = string_eval(input[index:index + length])
        a = data['value']
        b = a


def float_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        value = utils.tuple_to_string(input[index:index + length])
        data['value'] = float(value)
        a = data['value']
        b = a


def int_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        value = utils.tuple_to_string(input[index:index + length])
        data['value'] = int(value)


def true_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = True


def false_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = False


def null_value(state, input, index, length, data):
    if(state == id.SEM_PRE):
        data['value'] = None



# Copyright (c) 2022 Lowell D. Thomas, all rights reserved
# BSD-2-Clause (https://opensource.org/licenses/BSD-2-Clause)
#

# SUMMARY
#      rules = 61
#       udts = 4
#    opcodes = 292
#        ---   ABNF original opcodes
#        ALT = 37
#        CAT = 25
#        REP = 31
#        RNM = 95
#        TLS = 15
#        TBS = 49
#        TRG = 32
#        ---   SABNF super set opcodes
#        UDT = 4
#        AND = 2
#        NOT = 0
#        BKA = 0
#        BKN = 0
#        BKR = 0
#        ABG = 0
#        AEN = 2
# characters = [9 - 126] + user defined
#

# RULES
rules = ({'name': 'ini-file',
  'lower': 'ini-file',
  'index': 0,
  'line': 42,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 5)},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 1, 'children': (3, 4)},
              {'type': 4, 'index': 49},
              {'type': 4, 'index': 5},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 4, 'index': 1})},
 {'name': 'section',
  'lower': 'section',
  'index': 1,
  'line': 43,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2)},
              {'type': 4, 'index': 2},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 1, 'children': (4, 5)},
              {'type': 4, 'index': 49},
              {'type': 4, 'index': 5})},
 {'name': 'section-line',
  'lower': 'section-line',
  'index': 2,
  'line': 44,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 4, 'index': 3},
              {'type': 4, 'index': 4})},
 {'name': 'good-section-line',
  'lower': 'good-section-line',
  'index': 3,
  'line': 45,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 3, 4, 5, 6, 7, 9)},
              {'type': 6, 'string': (91,)},
              {'type': 4, 'index': 56},
              {'type': 4, 'index': 11},
              {'type': 4, 'index': 56},
              {'type': 6, 'string': (93,)},
              {'type': 4, 'index': 56},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 55},
              {'type': 4, 'index': 52})},
 {'name': 'bad-section-line',
  'lower': 'bad-section-line',
  'index': 4,
  'line': 46,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 4)},
              {'type': 6, 'string': (91,)},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 4, 'index': 60},
              {'type': 4, 'index': 52})},
 {'name': 'value-line',
  'lower': 'value-line',
  'index': 5,
  'line': 47,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 4, 'index': 6},
              {'type': 4, 'index': 8})},
 {'name': 'good-value',
  'lower': 'good-value',
  'index': 6,
  'line': 48,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 7, 8, 10)},
              {'type': 4, 'index': 12},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 2, 'children': (4, 5)},
              {'type': 4, 'index': 7},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 9},
              {'type': 4, 'index': 56},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 55},
              {'type': 4, 'index': 52})},
 {'name': 'key-delim',
  'lower': 'key-delim',
  'index': 7,
  'line': 49,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 7)},
              {'type': 2, 'children': (2, 3, 6)},
              {'type': 4, 'index': 56},
              {'type': 1, 'children': (4, 5)},
              {'type': 6, 'string': (61,)},
              {'type': 6, 'string': (58,)},
              {'type': 4, 'index': 56},
              {'type': 4, 'index': 57})},
 {'name': 'bad-value-line',
  'lower': 'bad-value-line',
  'index': 8,
  'line': 50,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 4, 8)},
              {'type': 1, 'children': (2, 3)},
              {'type': 5, 'min': 33, 'max': 90},
              {'type': 5, 'min': 92, 'max': 126},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 1, 'children': (6, 7)},
              {'type': 4, 'index': 54},
              {'type': 4, 'index': 60},
              {'type': 4, 'index': 52})},
 {'name': 'value-array',
  'lower': 'value-array',
  'index': 9,
  'line': 51,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2)},
              {'type': 4, 'index': 14},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 2, 'children': (4, 5)},
              {'type': 4, 'index': 10},
              {'type': 4, 'index': 14})},
 {'name': 'value-delim',
  'lower': 'value-delim',
  'index': 10,
  'line': 52,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 5)},
              {'type': 2, 'children': (2, 3, 4)},
              {'type': 4, 'index': 56},
              {'type': 6, 'string': (44,)},
              {'type': 4, 'index': 56},
              {'type': 4, 'index': 57})},
 {'name': 'section-name',
  'lower': 'section-name',
  'index': 11,
  'line': 53,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 4, 'index': 13},)},
 {'name': 'key-name',
  'lower': 'key-name',
  'index': 12,
  'line': 54,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 4, 'index': 13},)},
 {'name': 'name',
  'lower': 'name',
  'index': 13,
  'line': 55,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12)},
              {'type': 5, 'min': 97, 'max': 122},
              {'type': 5, 'min': 65, 'max': 90},
              {'type': 5, 'min': 48, 'max': 57},
              {'type': 6, 'string': (33,)},
              {'type': 5, 'min': 36, 'max': 38},
              {'type': 5, 'min': 40, 'max': 43},
              {'type': 5, 'min': 45, 'max': 46},
              {'type': 6, 'string': (60,)},
              {'type': 5, 'min': 62, 'max': 64},
              {'type': 5, 'min': 94, 'max': 95},
              {'type': 5, 'min': 123, 'max': 126})},
 {'name': 'value',
  'lower': 'value',
  'index': 14,
  'line': 66,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3, 4, 5)},
              {'type': 4, 'index': 19},
              {'type': 4, 'index': 15},
              {'type': 4, 'index': 31},
              {'type': 4, 'index': 33},
              {'type': 4, 'index': 30})},
 {'name': 'boolean',
  'lower': 'boolean',
  'index': 15,
  'line': 67,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3)},
              {'type': 4, 'index': 18},
              {'type': 4, 'index': 16},
              {'type': 4, 'index': 17})},
 {'name': 'true',
  'lower': 'true',
  'index': 16,
  'line': 68,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3)},
              {'type': 7, 'string': (116, 114, 117, 101)},
              {'type': 7, 'string': (121, 101, 115)},
              {'type': 7, 'string': (111, 110)})},
 {'name': 'false',
  'lower': 'false',
  'index': 17,
  'line': 69,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3)},
              {'type': 7, 'string': (102, 97, 108, 115, 101)},
              {'type': 7, 'string': (110, 111)},
              {'type': 7, 'string': (111, 102, 102)})},
 {'name': 'null',
  'lower': 'null',
  'index': 18,
  'line': 70,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3)},
              {'type': 7, 'string': (110, 117, 108, 108)},
              {'type': 7, 'string': (118, 111, 105, 100)},
              {'type': 7, 'string': (110, 111, 110, 101)})},
 {'name': 'number',
  'lower': 'number',
  'index': 19,
  'line': 71,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 4)},
              {'type': 1, 'children': (2, 3)},
              {'type': 4, 'index': 20},
              {'type': 4, 'index': 21},
              {'type': 12},
              {'type': 1, 'children': (6, 7, 8)},
              {'type': 6, 'string': (44,)},
              {'type': 4, 'index': 57},
              {'type': 4, 'index': 53})},
 {'name': 'int',
  'lower': 'int',
  'index': 20,
  'line': 74,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 5, 7)},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 1, 'children': (3, 4)},
              {'type': 6, 'string': (43,)},
              {'type': 6, 'string': (45,)},
              {'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 4, 'index': 59},
              {'type': 12},
              {'type': 1, 'children': (9, 10, 11)},
              {'type': 6, 'string': (44,)},
              {'type': 4, 'index': 57},
              {'type': 4, 'index': 53})},
 {'name': 'float',
  'lower': 'float',
  'index': 21,
  'line': 75,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 3, 4)},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 22},
              {'type': 4, 'index': 23},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 27})},
 {'name': 'sign',
  'lower': 'sign',
  'index': 22,
  'line': 76,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 7, 'string': (43,)},
              {'type': 7, 'string': (45,)})},
 {'name': 'decimal',
  'lower': 'decimal',
  'index': 23,
  'line': 77,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 8)},
              {'type': 2, 'children': (2, 3)},
              {'type': 4, 'index': 24},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 2, 'children': (5, 6)},
              {'type': 4, 'index': 25},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 26},
              {'type': 2, 'children': (9, 10)},
              {'type': 4, 'index': 25},
              {'type': 4, 'index': 26})},
 {'name': 'integer',
  'lower': 'integer',
  'index': 24,
  'line': 79,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 5, 'min': 48, 'max': 57})},
 {'name': 'dot',
  'lower': 'dot',
  'index': 25,
  'line': 80,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 7, 'string': (46,)},)},
 {'name': 'fraction',
  'lower': 'fraction',
  'index': 26,
  'line': 81,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 5, 'min': 48, 'max': 57})},
 {'name': 'exponent',
  'lower': 'exponent',
  'index': 27,
  'line': 82,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 4)},
              {'type': 7, 'string': (101,)},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 28},
              {'type': 4, 'index': 29})},
 {'name': 'esign',
  'lower': 'esign',
  'index': 28,
  'line': 83,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 7, 'string': (43,)},
              {'type': 7, 'string': (45,)})},
 {'name': 'exp',
  'lower': 'exp',
  'index': 29,
  'line': 84,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 5, 'min': 48, 'max': 57})},
 {'name': 'string',
  'lower': 'string',
  'index': 30,
  'line': 85,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4, 5, 6, 7, 8, 9, 10)},
              {'type': 6, 'string': (33,)},
              {'type': 5, 'min': 36, 'max': 38},
              {'type': 5, 'min': 40, 'max': 43},
              {'type': 5, 'min': 45, 'max': 46},
              {'type': 5, 'min': 48, 'max': 57},
              {'type': 6, 'string': (60,)},
              {'type': 5, 'min': 62, 'max': 91},
              {'type': 5, 'min': 93, 'max': 126},
              {'type': 4, 'index': 35})},
 {'name': 'd-quoted-string',
  'lower': 'd-quoted-string',
  'index': 31,
  'line': 87,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 3)},
              {'type': 6, 'string': (34,)},
              {'type': 4, 'index': 32},
              {'type': 6, 'string': (34,)})},
 {'name': 'd-quoted-value',
  'lower': 'd-quoted-value',
  'index': 32,
  'line': 88,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4, 5)},
              {'type': 5, 'min': 32, 'max': 33},
              {'type': 5, 'min': 35, 'max': 91},
              {'type': 5, 'min': 93, 'max': 126},
              {'type': 4, 'index': 35})},
 {'name': 's-quoted-string',
  'lower': 's-quoted-string',
  'index': 33,
  'line': 89,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 3)},
              {'type': 6, 'string': (39,)},
              {'type': 4, 'index': 34},
              {'type': 6, 'string': (39,)})},
 {'name': 's-quoted-value',
  'lower': 's-quoted-value',
  'index': 34,
  'line': 90,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4, 5)},
              {'type': 5, 'min': 32, 'max': 38},
              {'type': 5, 'min': 40, 'max': 91},
              {'type': 5, 'min': 93, 'max': 126},
              {'type': 4, 'index': 35})},
 {'name': 'escaped',
  'lower': 'escaped',
  'index': 35,
  'line': 91,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2)},
              {'type': 4, 'index': 36},
              {'type': 1,
               'children': (3,
                            4,
                            5,
                            6,
                            7,
                            8,
                            9,
                            10,
                            11,
                            12,
                            13,
                            14,
                            15,
                            16,
                            17,
                            18,
                            19)},
              {'type': 4, 'index': 36},
              {'type': 4, 'index': 41},
              {'type': 4, 'index': 37},
              {'type': 4, 'index': 38},
              {'type': 4, 'index': 39},
              {'type': 4, 'index': 40},
              {'type': 4, 'index': 42},
              {'type': 4, 'index': 43},
              {'type': 4, 'index': 44},
              {'type': 4, 'index': 45},
              {'type': 4, 'index': 46},
              {'type': 4, 'index': 47},
              {'type': 4, 'index': 48},
              {'type': 11, 'empty': False, 'index': 0},
              {'type': 11, 'empty': False, 'index': 1},
              {'type': 11, 'empty': False, 'index': 2},
              {'type': 11, 'empty': False, 'index': 3})},
 {'name': 'reverse-solidus',
  'lower': 'reverse-solidus',
  'index': 36,
  'line': 113,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (92,)},)},
 {'name': 'double-quote',
  'lower': 'double-quote',
  'index': 37,
  'line': 114,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (34,)},)},
 {'name': 'hash',
  'lower': 'hash',
  'index': 38,
  'line': 115,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (35,)},)},
 {'name': 'single-quote',
  'lower': 'single-quote',
  'index': 39,
  'line': 116,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (39,)},)},
 {'name': 'comma',
  'lower': 'comma',
  'index': 40,
  'line': 117,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (44,)},)},
 {'name': 'solidus',
  'lower': 'solidus',
  'index': 41,
  'line': 118,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (47,)},)},
 {'name': 'colon',
  'lower': 'colon',
  'index': 42,
  'line': 119,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (58,)},)},
 {'name': 'semicolon',
  'lower': 'semicolon',
  'index': 43,
  'line': 120,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (59,)},)},
 {'name': 'equal-sign',
  'lower': 'equal-sign',
  'index': 44,
  'line': 121,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (61,)},)},
 {'name': 'blank',
  'lower': 'blank',
  'index': 45,
  'line': 122,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (98,)},)},
 {'name': 'tab',
  'lower': 'tab',
  'index': 46,
  'line': 123,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (116,)},)},
 {'name': 'line-feed',
  'lower': 'line-feed',
  'index': 47,
  'line': 124,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (110,)},)},
 {'name': 'carriage-return',
  'lower': 'carriage-return',
  'index': 48,
  'line': 125,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 6, 'string': (114,)},)},
 {'name': 'blank-line',
  'lower': 'blank-line',
  'index': 49,
  'line': 126,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 4, 'index': 50},
              {'type': 4, 'index': 51})},
 {'name': 'good-blank-line',
  'lower': 'good-blank-line',
  'index': 50,
  'line': 127,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 4)},
              {'type': 4, 'index': 56},
              {'type': 3, 'min': 0, 'max': 1},
              {'type': 4, 'index': 55},
              {'type': 4, 'index': 52})},
 {'name': 'bad-blank-line',
  'lower': 'bad-blank-line',
  'index': 51,
  'line': 128,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 4, 5, 8, 10)},
              {'type': 1, 'children': (2, 3)},
              {'type': 6, 'string': (32,)},
              {'type': 6, 'string': (9,)},
              {'type': 4, 'index': 56},
              {'type': 1, 'children': (6, 7)},
              {'type': 5, 'min': 33, 'max': 58},
              {'type': 5, 'min': 60, 'max': 126},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 4, 'index': 60},
              {'type': 4, 'index': 52})},
 {'name': 'line-end',
  'lower': 'line-end',
  'index': 52,
  'line': 129,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3, 4)},
              {'type': 6, 'string': (13, 10)},
              {'type': 6, 'string': (10,)},
              {'type': 6, 'string': (13,)},
              {'type': 18})},
 {'name': 'digit-line-end',
  'lower': 'digit-line-end',
  'index': 53,
  'line': 130,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2, 3, 4)},
              {'type': 6, 'string': (13, 10)},
              {'type': 6, 'string': (10,)},
              {'type': 6, 'string': (13,)},
              {'type': 18})},
 {'name': 'line-continue',
  'lower': 'line-continue',
  'index': 54,
  'line': 131,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 2, 4, 5)},
              {'type': 6, 'string': (47,)},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 4, 'index': 60},
              {'type': 4, 'index': 52},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 1, 'children': (7, 8)},
              {'type': 6, 'string': (32,)},
              {'type': 6, 'string': (9,)})},
 {'name': 'comment',
  'lower': 'comment',
  'index': 55,
  'line': 132,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 2, 'children': (1, 4)},
              {'type': 1, 'children': (2, 3)},
              {'type': 6, 'string': (35,)},
              {'type': 6, 'string': (59,)},
              {'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 4, 'index': 60})},
 {'name': 'owsp',
  'lower': 'owsp',
  'index': 56,
  'line': 133,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 0, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4)},
              {'type': 6, 'string': (32,)},
              {'type': 6, 'string': (9,)},
              {'type': 4, 'index': 54})},
 {'name': 'wsp',
  'lower': 'wsp',
  'index': 57,
  'line': 134,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 3, 'min': 1, 'max': 9223372036854775807},
              {'type': 1, 'children': (2, 3, 4)},
              {'type': 6, 'string': (32,)},
              {'type': 6, 'string': (9,)},
              {'type': 4, 'index': 54})},
 {'name': 'alpha',
  'lower': 'alpha',
  'index': 58,
  'line': 135,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 5, 'min': 65, 'max': 90},
              {'type': 5, 'min': 97, 'max': 122})},
 {'name': 'digit',
  'lower': 'digit',
  'index': 59,
  'line': 136,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 5, 'min': 48, 'max': 57},)},
 {'name': 'any',
  'lower': 'any',
  'index': 60,
  'line': 137,
  'is_bkru': False,
  'is_bkrr': False,
  'has_bkrr': False,
  'opcodes': ({'type': 1, 'children': (1, 2)},
              {'type': 5, 'min': 32, 'max': 126},
              {'type': 6, 'string': (9,)})})

# UDTS
udts = ({'name': 'u_unicode8',
  'lower': 'u_unicode8',
  'index': 0,
  'is_bkru': False,
  'is_bkrr': False,
  'empty': False},
 {'name': 'u_unicode4',
  'lower': 'u_unicode4',
  'index': 1,
  'is_bkru': False,
  'is_bkrr': False,
  'empty': False},
 {'name': 'u_hexadecimal',
  'lower': 'u_hexadecimal',
  'index': 2,
  'is_bkru': False,
  'is_bkrr': False,
  'empty': False},
 {'name': 'u_escaped-error',
  'lower': 'u_escaped-error',
  'index': 3,
  'is_bkru': False,
  'is_bkrr': False,
  'empty': False})

has_bkru = False
has_bkrr = False


def to_string():
    '''Displays the original SABNF syntax.'''
    sabnf = ""
    sabnf += ";\n"
    sabnf += "; Ref: https://en.wikipedia.org/wiki/INI_file\n"
    sabnf += "; This SABNF grammar encompasses most of the stable and varying feature of the WikiPedia\n"
    sabnf += "; artcle. The only significant ommission is that subsections are not supported.\n"
    sabnf += ";\n"
    sabnf += "; comments begin with the semicolon(;) or hash(#) and continue to the end of the line\n"
    sabnf += "; comments may appear on valid section and value lines as well as blank lines\n"
    sabnf += "; line ends may be CRLF, LF or CR\n"
    sabnf += "; lines may be continued by ending them with a backslash and the following line begins with space or tab\n"
    sabnf += "; e.g.\n"
    sabnf += "; key = value , value, \\\n"
    sabnf += ";       value\n"
    sabnf += ";\n"
    sabnf += "; section and key names are case sensitive\n"
    sabnf += "; section and key names are liberally defined as one or more characters from the set:\n"
    sabnf += ";   a-z\n"
    sabnf += ";   A-Z\n"
    sabnf += ";   0-9\n"
    sabnf += ";   !$%&()*+-.<>?@^_`{|}~\n"
    sabnf += ";   (basically, alphabet, numeric, underscore, hyphen or any character not used as a special or reserved character)\n"
    sabnf += ";\n"
    sabnf += "; keys may have multiple values\n"
    sabnf += ";   - multiple values may be given as a comma or space delimited list on a single line\n"
    sabnf += ";   - values are \"disjoint\" - values listed for multiple occurrances of a key are concatenated\n"
    sabnf += ";   - values may be numbers, booleans, or strings\n"
    sabnf += ";   - strings may be single-quoted, double-quoted strings or unquoted\n"
    sabnf += ";\n"
    sabnf += "; keys appearing before the first section are allowed and termed the \"global\" section.\n"
    sabnf += ";\n"
    sabnf += "; section names are optional - the file may consist of only a \"global\" section\n"
    sabnf += ";\n"
    sabnf += "; sections are \"disjoint\",\n"
    sabnf += ";   - that is the keys in multiple occurrences of a section name are\n"
    sabnf += ";     joined together as if they appeared contiguously in a single section\n"
    sabnf += ";\n"
    sabnf += "; sections end at the beginning of a new section or the end of file\n"
    sabnf += ";\n"
    sabnf += "; The grammar is designed to accept any string of ASCII characters without failure.\n"
    sabnf += "; The \"error productions\", bad-section-line, bad-value-line, bad-blank-line are meant to accept all lines\n"
    sabnf += "; that are not otherwise correct blank, section or value lines. This allows the\n"
    sabnf += "; parser to report the input errors in an application-dependent manner.\n"
    sabnf += ";\n"
    sabnf += "ini-file            = *(blank-line / value-line) *section\n"
    sabnf += "section             = section-line *(blank-line / value-line)\n"
    sabnf += "section-line        = good-section-line / bad-section-line\n"
    sabnf += "good-section-line   = %d91 owsp section-name owsp %d93 owsp [comment] line-end\n"
    sabnf += "bad-section-line    = %d91 *any line-end;\n"
    sabnf += "value-line          = good-value / bad-value-line\n"
    sabnf += "good-value          = key-name [key-delim [value-array]] owsp [comment] line-end\n"
    sabnf += "key-delim           = (owsp (%d61 / %d58) owsp) / wsp ; = or : or white space\n"
    sabnf += "bad-value-line      = (%d33-90 / %d92-126) *(line-continue / any) line-end\n"
    sabnf += "value-array         = value *(value-delim value)\n"
    sabnf += "value-delim         = (owsp %d44 owsp) / wsp\n"
    sabnf += "section-name        = name\n"
    sabnf += "key-name            = name\n"
    sabnf += "name                = 1*(%d97-122 \n"
    sabnf += "                    / %d65-90 \n"
    sabnf += "                    / %d48-57 \n"
    sabnf += "                    / %d33 \n"
    sabnf += "                    / %d36-38\n"
    sabnf += "                    / %d40-43\n"
    sabnf += "                    / %d45-46\n"
    sabnf += "                    / %d60 \n"
    sabnf += "                    / %d62-64 \n"
    sabnf += "                    / %d94-95\n"
    sabnf += "                    / %d123-126)\n"
    sabnf += "value               = number / boolean / d-quoted-string / s-quoted-string / string\n"
    sabnf += "boolean             = null / true / false\n"
    sabnf += "true                = \"true\" / \"yes\" / \"on\"\n"
    sabnf += "false               = \"false\" / \"no\" / \"off\"\n"
    sabnf += "null                = \"null\" / \"void\" / \"none\"\n"
    sabnf += "number              = (int / float) &(%d44 / wsp / digit-line-end)\n"
    sabnf += "                      ; digit-line-end is a line that does not increment the line counter\n"
    sabnf += "                      ; numbers must be followed by a delimiter or line end\n"
    sabnf += "int                 = [%d43 / %d45] 1*digit &(%d44 / wsp / digit-line-end)\n"
    sabnf += "float               = [sign] decimal [exponent]\n"
    sabnf += "sign                = \"+\" / \"-\"\n"
    sabnf += "decimal             = integer [dot [fraction]]\n"
    sabnf += "                    / dot fraction\n"
    sabnf += "integer             = 1*%d48-57\n"
    sabnf += "dot                 = \".\"\n"
    sabnf += "fraction            = 1*%d48-57\n"
    sabnf += "exponent            = \"e\" [esign] exp\n"
    sabnf += "esign               = \"+\" / \"-\"\n"
    sabnf += "exp                 = 1*%d48-57                      \n"
    sabnf += "string              = 1*(%d33 / %d36-38 / %d40-43 / %d45-46 / %d48-57 / %d60 / %d62-91 / %d93-126 / escaped)\n"
    sabnf += "                      ; characters not allowed: space \" # ' , / : ; = \\\n"
    sabnf += "d-quoted-string     = %d34  d-quoted-value %d34\n"
    sabnf += "d-quoted-value      = 1*(%d32-33 / %d35-91 / %d93-126 / escaped)\n"
    sabnf += "s-quoted-string     = %d39 s-quoted-value %d39\n"
    sabnf += "s-quoted-value      = 1*(%d32-38  /  %d40-91 / %d93-126 / escaped)\n"
    sabnf += "escaped             = reverse-solidus (reverse-solidus\n"
    sabnf += "                    / solidus\n"
    sabnf += "                    / double-quote\n"
    sabnf += "                    / hash\n"
    sabnf += "                    / single-quote\n"
    sabnf += "                    / comma\n"
    sabnf += "                    / colon\n"
    sabnf += "                    / semicolon\n"
    sabnf += "                    / equal-sign\n"
    sabnf += "                    / blank\n"
    sabnf += "                    / tab\n"
    sabnf += "                    / line-feed\n"
    sabnf += "                    / carriage-return\n"
    sabnf += "                    / u_unicode8\n"
    sabnf += "                    / u_unicode4\n"
    sabnf += "                    / u_hexadecimal\n"
    sabnf += "                    / u_escaped-error)\n"
    sabnf += "; u_hexadecimal - UDT \"xhh\", exactly 2 hex digits [0-9a-fA-F]                    \n"
    sabnf += "; u_unicode4    - UDT \"uhhhh\", exactly 4 hex digits and semantic checking for valid Unicode characters\n"
    sabnf += ";                 surrogates, xd800 - xdfff, not allowed\n"
    sabnf += "; u_unicode8    - UDT \"Uhhhhhhhh\", exactly 8 hex digits and semantic checking for valid Unicode characters\n"
    sabnf += ";                 surrogates, xd800 - xdfff, and > x10ffff not allowed\n"
    sabnf += "reverse-solidus     = %x5c\n"
    sabnf += "double-quote        = %x22\n"
    sabnf += "hash                = %x23\n"
    sabnf += "single-quote        = %x27\n"
    sabnf += "comma               = %x2c\n"
    sabnf += "solidus             = %x2f\n"
    sabnf += "colon               = %x3a\n"
    sabnf += "semicolon           = %x3b\n"
    sabnf += "equal-sign          = %x3d\n"
    sabnf += "blank               = %s\"b\" ; %x20\n"
    sabnf += "tab                 = %s\"t\" ; %x09\n"
    sabnf += "line-feed           = %s\"n\" ; %x0a\n"
    sabnf += "carriage-return     = %s\"r\" ; %x0d\n"
    sabnf += "blank-line          = good-blank-line / bad-blank-line\n"
    sabnf += "good-blank-line     = owsp [comment] line-end\n"
    sabnf += "bad-blank-line      = (%d32 / %d9) owsp (%d33-58 / %d60-126) *any line-end\n"
    sabnf += "line-end            = %d13.10 / %d10 / %d13 / %$\n"
    sabnf += "digit-line-end      = %d13.10 / %d10 / %d13 / %$\n"
    sabnf += "line-continue       = %d47 *any line-end *(%d32 / %d9)\n"
    sabnf += "comment             = (%d35 / %d59) *any\n"
    sabnf += "owsp                = *(%d32 / %d9 / line-continue)\n"
    sabnf += "wsp                 = 1*(%d32 / %d9 / line-continue)\n"
    sabnf += "alpha               = %d65-90 / %d97-122\n"
    sabnf += "digit               = %d48-57\n"
    sabnf += "any                 = %d32-126 / %d9\n"
    return sabnf




import sys
import os
# add the current working directory to the path
# DO NOT MOVE THE FOLLOWING STATEMENT
# if using autopep8 formatter, for example, set argument '--ignore=E402'
sys.path.append(os.getcwd())
import unittest
from python_ini.ini_writer import IniWriter


class TestExceptions(unittest.TestCase):
    """Test all raised exceptions."""

    def test_comments(self):
        '''Normalization of comments execeptions.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.comment(10)
        # print()
        # print(ctx.exception)
        self.assertIn('ASCII', str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.comment('a\xffb')
        # print(ctx.exception)
        self.assertIn('char code(255)', str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.comment('a\tb')
        # print(ctx.exception)
        self.assertIn('char code(9)', str(ctx.exception))

    def test_strings(self):
        '''Key string values.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.key('name', 'a\ud800b')
        # print()
        # print(ctx.exception)
        self.assertIn('string has Unicode surrogate value', str(ctx.exception))

    def test_booleans(self):
        '''Set boolean values.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.booleans(true='TRUE')
            w.booleans(true='on')
            w.booleans(true='YeS')
            w.booleans(true='yep')
        # print()
        # print(ctx.exception)
        self.assertIn('yep', str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.booleans(false='FALSE')
            w.booleans(false='no')
            w.booleans(false='oFF')
            w.booleans(false='nope')
        # print(ctx.exception)
        self.assertIn('nope', str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.booleans(none='NONE')
            w.booleans(none='null')
            w.booleans(none='VoId')
            w.booleans(none='nill')
        # print(ctx.exception)
        self.assertIn('nill', str(ctx.exception))

    def test_comment_tab(self):
        '''Comment tab values.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.comment_tab(-10)
        # print()
        # print(ctx.exception)
        self.assertIn('argument must be positive integer', str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.comment_tab('10')
        # print(ctx.exception)
        self.assertIn('argument must be positive integer', str(ctx.exception))

    def test_delimiters(self):
        '''Setting delimiter values.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.delimiters(comment=10)
        # print()
        # print(ctx.exception)
        self.assertIn(
            'comment argument must be semicolon(;) or hash(#)',
            str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.delimiters(comment=':')
        # print(ctx.exception)
        self.assertIn(
            'comment argument must be semicolon(;) or hash(#)',
            str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.delimiters(key='#')
        # print(ctx.exception)
        self.assertIn(
            'key argument must be equals(=), comma(,) or space',
            str(ctx.exception))
        with self.assertRaises(Exception) as ctx:
            w.delimiters(value='#')
        # print(ctx.exception)
        self.assertIn(
            'value argument must be comma(,) or space',
            str(ctx.exception))

    def test_key_values(self):
        '''key values.'''
        w = IniWriter()
        with self.assertRaises(Exception) as ctx:
            w.key('name', {})
        # print()
        # print(ctx.exception)
        self.assertIn(
            'invalid value',
            str(ctx.exception))


if __name__ == '__main__':
    unittest.main()



import unittest
from python_ini.ini_file import IniFile


def ini_data(fname):
    with open(fname, 'r') as fd:
        input = fd.read()
        print()
        print('ini data: ' + fname)
        print(input)


class TestGlobalSingle(unittest.TestCase):
    """Test parsing only global keys in single-value mode."""

    def test_globals1(self):
        '''Simple key value pairs.'''
        fname = 'tests/data/simple_global.ini'
        # ini_data(fname)
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        self.assertTrue(keys[0] == 'home_page')
        self.assertTrue(keys[1] == 'email')
        self.assertTrue(keys[2] == 'you')
        self.assertTrue(keys[3] == 'them')
        value = ini.get_values('home_page')
        self.assertTrue(value == 'https://my.homepage.com')
        value = ini.get_values('email')
        self.assertTrue(value == 'me@ini.com')
        value = ini.get_values('you')
        self.assertTrue(value == 'you@your.com')
        value = ini.get_values('them')
        self.assertTrue(value == 'them@their.com')

    def test_globals2(self):
        '''True flags in the global section'''
        fname = 'tests/data/true_flag_globals.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        # print()
        # print(keys)
        for key in keys:
            value = ini.get_values(key)
            # print()
            # print(key)
            # print(value)
            self.assertTrue(value)

    def test_globals3(self):
        '''True disjoint section and overriding multiple key names'''
        fname = 'tests/data/sections.ini'
        ini = IniFile()
        ini.parse(fname)
        sections = ini.get_sections()
        self.assertTrue(sections[0] == 'SECTION1')
        self.assertTrue(sections[1] == '_SECTION_')
        self.assertTrue(len(sections) == 2)
        keys = ini.get_section_keys(sections[0])
        self.assertTrue(len(keys) == 4)
        # first section values
        self.assertTrue(sections[0] == 'SECTION1')
        value = ini.get_section_values(sections[0], keys[0])
        self.assertTrue(value == 5)  # override of first value for this key
        value = ini.get_section_values(sections[0], keys[1])
        self.assertTrue(value == 2)
        value = ini.get_section_values(sections[0], keys[2])
        self.assertTrue(value == 3)
        value = ini.get_section_values(sections[0], keys[3])
        self.assertTrue(value == 4)
        # second section values
        keys = ini.get_section_keys(sections[1])
        self.assertTrue(sections[1] == '_SECTION_')
        value = ini.get_section_values(sections[1], keys[0])
        # print()
        # print(keys[0])
        # print(value)
        self.assertTrue(value == 50)  # override of first value for this key
        value = ini.get_section_values(sections[1], keys[1])
        self.assertTrue(value == 20)
        value = ini.get_section_values(sections[1], keys[2])
        self.assertTrue(value == 30)
        value = ini.get_section_values(sections[1], keys[3])
        self.assertTrue(value == 40)

    def test_globals4(self):
        '''Section and key names with all special characters.'''
        inistr = '''[SECTION_section_0123456789_!$%()*+-.<>?@^_{|}~]
KEY.key_10123456789_!$%()*+-.<>?@^_{|}~ = value
'''
        ini = IniFile()
        ini.parse(fstr=inistr)
        sections = ini.get_sections()
        self.assertTrue(
            sections[0] == 'SECTION_section_0123456789_!$%()*+-.<>?@^_{|}~')
        keys = ini.get_section_keys(sections[0])
        self.assertTrue(
            keys[0] == 'KEY.key_10123456789_!$%()*+-.<>?@^_{|}~')
        value = ini.get_section_values(sections[0], keys[0])
        self.assertTrue(value == 'value')

    def test_globals5(self):
        '''Quoted strings.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        # print()
        # print(keys)
        value = ini.get_values(keys[0])
        # print(value)
        self.assertTrue(value == 'my string, "baby"')
        value = ini.get_values(keys[1])
        # print(value)
        self.assertTrue(value == "double down, 'baby'")

    def test_globals6(self):
        '''Escaped characters.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        value = ini.get_values(keys[2])
        self.assertTrue(value[0] == 'a')
        self.assertTrue(value[1] == 'B')
        self.assertTrue(value[2] == '3')
        self.assertTrue(value[3] == '\\')
        self.assertTrue(value[4] == '"')
        self.assertTrue(value[5] == '#')
        self.assertTrue(value[6] == "'")
        self.assertTrue(value[7] == ',')
        self.assertTrue(value[8] == '/')
        self.assertTrue(value[9] == ':')
        self.assertTrue(value[10] == ';')
        self.assertTrue(value[11] == '=')
        self.assertTrue(value[12] == ' ')
        self.assertTrue(ord(value[13]) == 9)
        self.assertTrue(ord(value[14]) == 10)
        self.assertTrue(ord(value[15]) == 13)
        self.assertTrue(ord(value[16]) == 0xff)
        self.assertTrue(ord(value[17]) == 0x12ab)

    def test_globals7(self):
        '''Escaped characters in single-quoted string.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        value = ini.get_values(keys[3])
        self.assertTrue(value[0] == 'a')
        self.assertTrue(value[1] == 'B')
        self.assertTrue(value[2] == '3')
        self.assertTrue(value[3] == '\\')
        self.assertTrue(value[4] == '"')
        self.assertTrue(value[5] == '#')
        self.assertTrue(value[6] == "'")
        self.assertTrue(value[7] == ',')
        self.assertTrue(value[8] == '/')
        self.assertTrue(value[9] == ':')
        self.assertTrue(value[10] == ';')
        self.assertTrue(value[11] == '=')
        self.assertTrue(value[12] == ' ')
        self.assertTrue(ord(value[13]) == 9)
        self.assertTrue(ord(value[14]) == 10)
        self.assertTrue(ord(value[15]) == 13)
        self.assertTrue(ord(value[16]) == 0xff)
        self.assertTrue(ord(value[17]) == 0x12ab)

    def test_globals8(self):
        '''Escaped characters in double-quoted string.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        value = ini.get_values(keys[4])
        self.assertTrue(value[0] == 'a')
        self.assertTrue(value[1] == 'B')
        self.assertTrue(value[2] == '3')
        self.assertTrue(value[3] == '\\')
        self.assertTrue(value[4] == '"')
        self.assertTrue(value[5] == '#')
        self.assertTrue(value[6] == "'")
        self.assertTrue(value[7] == ',')
        self.assertTrue(value[8] == '/')
        self.assertTrue(value[9] == ':')
        self.assertTrue(value[10] == ';')
        self.assertTrue(value[11] == '=')
        self.assertTrue(value[12] == ' ')
        self.assertTrue(ord(value[13]) == 9)
        self.assertTrue(ord(value[14]) == 10)
        self.assertTrue(ord(value[15]) == 13)
        self.assertTrue(ord(value[16]) == 0xff)
        self.assertTrue(ord(value[17]) == 0x12ab)

    def test_globals9(self):
        '''Un-escaped characters in double-quoted string.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        value = ini.get_values(keys[5])
        self.assertTrue(value[0] == '#')
        self.assertTrue(value[1] == "'")
        self.assertTrue(value[2] == ',')
        self.assertTrue(value[3] == '/')
        self.assertTrue(value[4] == ':')
        self.assertTrue(value[5] == ';')
        self.assertTrue(value[6] == "=")
        self.assertTrue(value[7] == ' ')

    def test_globals10(self):
        '''Tortured line continues.'''
        fname = 'tests/data/quoted_strings.ini'
        ini = IniFile()
        ini.parse(fname)
        keys = ini.get_keys()
        value = ini.get_values(keys[6])
        self.assertTrue(value == 'double-continue')
        value = ini.get_values(keys[7])
        self.assertTrue(value == 'comment-continue')
        value = ini.get_values(keys[8])
        self.assertTrue(value == 'continue3')

    def test_globals11(self):
        '''Data from file handle.'''
        fname = 'tests/data/true_flag_globals.ini'
        with open(fname, 'r') as fd:
            ini = IniFile()
            ini.parse(fhandle=fd)
            keys = ini.get_keys()
            # print()
            # print(keys)
            for key in keys:
                value = ini.get_values(key)
                self.assertTrue(value)

    def test_globals12(self):
        '''Data from string.'''
        fname = 'tests/data/true_flag_globals.ini'
        with open(fname, 'r') as fd:
            input = fd.read()
            ini = IniFile()
            ini.parse(fstr=input)
            keys = ini.get_keys()
            # print()
            # print(keys)
            for key in keys:
                value = ini.get_values(key)
                self.assertTrue(value)


if __name__ == '__main__':
    unittest.main()
